/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Copyright (C) 1999-2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>  - Asterisk Author
 * Matthew Boehm <mboehm@cytelcom.com> - MySQL RealTime Driver Author
 *
 * res_config_mysql.c <mysql plugin for RealTime configuration engine>
 *
 * v2.0   - (10-07-05) - mutex_lock fixes (bug #4973, comment #0034602)
 *
 * v1.9   - (08-19-05) - Added support to correctly honor the family database specified
 *                       in extconfig.conf (bug #4973)
 *
 * v1.8   - (04-21-05) - Modified return values of update_mysql to better indicate
 *                       what really happened.
 *
 * v1.7   - (01-28-05) - Fixed non-initialization of ast_category struct
 *                       in realtime_multi_mysql function which caused segfault. 
 *
 * v1.6   - (00-00-00) - Skipped to bring comments into sync with version number in CVS.
 *
 * v1.5.1 - (01-26-05) - Added better(?) locking stuff
 *
 * v1.5   - (01-26-05) - Brought up to date with new config.h changes (bug #3406)
 *                     - Added in extra locking provided by georg (bug #3248)
 *
 * v1.4   - (12-02-04) - Added realtime_multi_mysql function
 *                        This function will return an ast_config with categories,
 *                        unlike standard realtime_mysql which only returns
 *                        a linked list of ast_variables
 *
 * v1.3   - (12-01-04) - Added support other operators
 *                       Ex: =, !=, LIKE, NOT LIKE, RLIKE, etc...
 *
 * v1.2   - (11-DD-04) - Added reload. Updated load and unload.
 *                       Code beautification (doc/CODING-GUIDELINES)
 */

/*** MODULEINFO
	<depend>mysqlclient</depend>
 ***/

#include <asterisk.h>

#include <asterisk/channel.h>
#include <asterisk/logger.h>
#include <asterisk/config.h>
#include <asterisk/module.h>
#include <asterisk/lock.h>
#include <asterisk/options.h>
#include <asterisk/cli.h>
#include <asterisk/utils.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdio.h>
#include <mysql/mysql.h>
#include <mysql/mysql_version.h>
#include <mysql/errmsg.h>

#define RES_CONFIG_MYSQL_CONF "res_mysql.conf"
#define	READHANDLE	0
#define	WRITEHANDLE	1

static struct mysql_conn {
	ast_mutex_t	lock;
	MYSQL       handle;
	char        host[50];
	char        name[50];
	char        user[50];
	char        pass[50];
	char        sock[50];
	int         port;
	int         connected;
	time_t      connect_time;
} dbread, dbwrite;

static int parse_config(void);
static int mysql_reconnect(struct mysql_conn *conn);
static char *handle_cli_realtime_mysql_status(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static int load_mysql_config(struct ast_config *config, const char *category, struct mysql_conn *conn);

static struct ast_cli_entry cli_realtime_mysql_status[] = {
	AST_CLI_DEFINE(handle_cli_realtime_mysql_status, "Shows connection information for the MySQL RealTime driver"),
};
/*
        { "realtime", "mysql", "status", NULL }, realtime_mysql_status,
        "Shows connection information for the MySQL RealTime driver", cli_realtime_mysql_status_usage, NULL };
*/
static struct ast_variable *realtime_mysql(const char *database, const char *table, va_list ap)
{
	MYSQL_RES *result;
	MYSQL_ROW row;
	MYSQL_FIELD *fields;
	int numFields, i, valsz;
	char sql[512];
	char buf[511]; /* Keep this size uneven as it is 2n+1. */
	char *stringp;
	char *chunk;
	char *op;
	char dbname[50] = "";
	const char *newparam, *newval;
	struct ast_variable *var=NULL, *prev=NULL;

	if (!table) {
		ast_log(LOG_WARNING, "MySQL RealTime: No table specified.\n");
		return NULL;
	}

	/* Get the first parameter and first value in our list of passed paramater/value pairs */
	newparam = va_arg(ap, const char *);
	newval = va_arg(ap, const char *);
	if (!newparam || !newval)  {
		ast_log(LOG_WARNING, "MySQL RealTime: Realtime retrieval requires at least 1 parameter and 1 value to search on.\n");
		return NULL;
	}

	/* Must connect to the server before anything else, as the escape function requires the mysql handle. */
	ast_mutex_lock(&dbread.lock);
	if (!mysql_reconnect(&dbread)) {
		ast_mutex_unlock(&dbread.lock);
		return NULL;
	}

	/* Create the first part of the query using the first parameter/value pairs we just extracted
	   If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat */

	if (!strchr(newparam, ' ')) 
		op = " ="; 
	else 
		op = "";

	if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
		valsz = (sizeof(buf) - 1) / 2;
	mysql_real_escape_string(&dbread.handle, buf, newval, valsz);
	if (strcmp(database, dbread.name)) {
		snprintf(dbname, sizeof(dbname), "%s.", database);
	}
	snprintf(sql, sizeof(sql), "SELECT * FROM %s%s WHERE %s%s '%s'", dbname, table, newparam, op, buf);
	while ((newparam = va_arg(ap, const char *))) {
		newval = va_arg(ap, const char *);
		if (!strchr(newparam, ' ')) 
			op = " ="; 
		else
			op = "";
		if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
			valsz = (sizeof(buf) - 1) / 2;
		mysql_real_escape_string(&dbread.handle, buf, newval, valsz);
		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " AND %s%s '%s'", newparam, op, buf);
	}
	va_end(ap);

	ast_debug(1, "MySQL RealTime: Retrieve SQL: %s\n", sql);

	/* Execution. */
	if (mysql_real_query(&dbread.handle, sql, strlen(sql))) {
		ast_log(LOG_WARNING, "MySQL RealTime: Failed to query database. Check debug for more info.\n");
		ast_debug(1, "MySQL RealTime: Query: %s\n", sql);
		ast_debug(1, "MySQL RealTime: Query Failed because: %s\n", mysql_error(&dbread.handle));
		ast_mutex_unlock(&dbread.lock);
		return NULL;
	}

	if ((result = mysql_store_result(&dbread.handle))) {
		numFields = mysql_num_fields(result);
		fields = mysql_fetch_fields(result);

		while ((row = mysql_fetch_row(result))) {
			for (i = 0; i < numFields; i++) {
				if (ast_strlen_zero(row[i]))
					continue;
				for (stringp = ast_strdupa(row[i]), chunk = strsep(&stringp, ";"); chunk; chunk = strsep(&stringp, ";")) {
					if (chunk && !ast_strlen_zero(ast_strip(chunk))) {
						if (prev) {
							prev->next = ast_variable_new(fields[i].name, chunk, "");
							if (prev->next) {
								prev = prev->next;
							}
						} else {
							prev = var = ast_variable_new(fields[i].name, chunk, "");
						}
					}
				}
			}
		}
	} else {
		ast_log(LOG_WARNING, "MySQL RealTime: Could not find any rows in table %s.\n", table);
	}

	ast_mutex_unlock(&dbread.lock);
	mysql_free_result(result);

	return var;
}

static struct ast_config *realtime_multi_mysql(const char *database, const char *table, va_list ap)
{
	MYSQL_RES *result;
	MYSQL_ROW row;
	MYSQL_FIELD *fields;
	int numFields, i, valsz;
	char sql[512];
	char buf[511]; /* Keep this size uneven as it is 2n+1. */
	const char *initfield = NULL;
	char *stringp;
	char *chunk;
	char *op;
	char dbname[50] = "";
	const char *newparam, *newval;
	struct ast_variable *var = NULL;
	struct ast_config *cfg = NULL;
	struct ast_category *cat = NULL;

	if (!table) {
		ast_log(LOG_WARNING, "MySQL RealTime: No table specified.\n");
		return NULL;
	}
	
	cfg = ast_config_new();
	if (!cfg) {
		/* If I can't alloc memory at this point, why bother doing anything else? */
		ast_log(LOG_WARNING, "Out of memory!\n");
		return NULL;
	}

	/* Get the first parameter and first value in our list of passed paramater/value pairs */
	newparam = va_arg(ap, const char *);
	newval = va_arg(ap, const char *);
	if (!newparam || !newval)  {
		ast_log(LOG_WARNING, "MySQL RealTime: Realtime retrieval requires at least 1 parameter and 1 value to search on.\n");
		ast_config_destroy(cfg);
		return NULL;
	}

	initfield = ast_strdupa(newparam);
	if (initfield && (op = strchr(initfield, ' '))) {
		*op = '\0';
	}

	/* Must connect to the server before anything else, as the escape function requires the mysql handle. */
	ast_mutex_lock(&dbread.lock);
	if (!mysql_reconnect(&dbread)) {
		ast_mutex_unlock(&dbread.lock);
		ast_config_destroy(cfg);
		return NULL;
	}

	/* Create the first part of the query using the first parameter/value pairs we just extracted
	   If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat */

	if (!strchr(newparam, ' '))
		op = " =";
	else
		op = "";

	if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
		valsz = (sizeof(buf) - 1) / 2;
	mysql_real_escape_string(&dbread.handle, buf, newval, valsz);
	if (strcmp(database, dbread.name)) {
		snprintf(dbname, sizeof(dbname), "%s.", database);
	}
	snprintf(sql, sizeof(sql), "SELECT * FROM %s%s WHERE %s%s '%s'", dbname, table, newparam, op, buf);
	while ((newparam = va_arg(ap, const char *))) {
		newval = va_arg(ap, const char *);
		if (!strchr(newparam, ' ')) op = " ="; else op = "";
		if ((valsz = strlen(newval)) * 2 + 1 > sizeof(buf))
			valsz = (sizeof(buf) - 1) / 2;
		mysql_real_escape_string(&dbread.handle, buf, newval, valsz);
		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " AND %s%s '%s'", newparam, op, buf);
	}

	if (initfield) {
		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " ORDER BY %s", initfield);
	}

	va_end(ap);

	ast_debug(1, "MySQL RealTime: Retrieve SQL: %s\n", sql);

	/* Execution. */
	if (mysql_real_query(&dbread.handle, sql, strlen(sql))) {
		ast_log(LOG_WARNING, "MySQL RealTime: Failed to query database. Check debug for more info.\n");
		ast_debug(1, "MySQL RealTime: Query: %s\n", sql);
		ast_debug(1, "MySQL RealTime: Query Failed because: %s\n", mysql_error(&dbread.handle));
		ast_mutex_unlock(&dbread.lock);
		ast_config_destroy(cfg);
		return NULL;
	}

	if ((result = mysql_store_result(&dbread.handle))) {
		numFields = mysql_num_fields(result);
		fields = mysql_fetch_fields(result);

		while ((row = mysql_fetch_row(result))) {
			var = NULL;
			cat = ast_category_new("", "", -1);
			if (!cat) {
				ast_log(LOG_WARNING, "Out of memory!\n");
				continue;
			}
			for (i = 0; i < numFields; i++) {
				if (ast_strlen_zero(row[i]))
					continue;
				for (stringp = ast_strdupa(row[i]), chunk = strsep(&stringp, ";"); chunk; chunk = strsep(&stringp, ";")) {
					if (chunk && !ast_strlen_zero(ast_strip(chunk))) {
						if (initfield && !strcmp(initfield, fields[i].name)) {
							ast_category_rename(cat, chunk);
						}
						var = ast_variable_new(fields[i].name, chunk, "");
						ast_variable_append(cat, var);
					}
				}
			}
			ast_category_append(cfg, cat);
		}
	} else {
		ast_log(LOG_WARNING, "MySQL RealTime: Could not find any rows in table %s.\n", table);
	}

	ast_mutex_unlock(&dbread.lock);
	mysql_free_result(result);

	return cfg;
}

static int update_mysql(const char *database, const char *table, const char *keyfield, const char *lookup, va_list ap)
{
	my_ulonglong numrows;
	char sql[512];
	char buf[511]; /* Keep this size uneven as it is 2n+1. */
	char dbname[50] = "";
	int valsz;
	const char *newparam, *newval;

	if (!table) {
		ast_log(LOG_WARNING, "MySQL RealTime: No table specified.\n");
		return -1;
	}

	/* Get the first parameter and first value in our list of passed paramater/value pairs */
	newparam = va_arg(ap, const char *);
	newval = va_arg(ap, const char *);
	if (!newparam || !newval)  {
		ast_log(LOG_WARNING, "MySQL RealTime: Realtime retrieval requires at least 1 parameter and 1 value to search on.\n");
               return -1;
	}

	/* Must connect to the server before anything else, as the escape function requires the mysql handle. */
	ast_mutex_lock(&dbwrite.lock);
	if (!mysql_reconnect(&dbwrite)) {
		ast_mutex_unlock(&dbwrite.lock);
		return -1;
	}

	/* Create the first part of the query using the first parameter/value pairs we just extracted
	   If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat */

	if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
		valsz = (sizeof(buf) - 1) / 2;
	mysql_real_escape_string(&dbwrite.handle, buf, newval, valsz);
	if (strcmp(database, dbwrite.name)) {
		snprintf(dbname, sizeof(dbname), "%s.", database);
	}
	snprintf(sql, sizeof(sql), "UPDATE %s%s SET %s = '%s'", dbname, table, newparam, buf);
	while((newparam = va_arg(ap, const char *))) {
		newval = va_arg(ap, const char *);
		if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
			valsz = (sizeof(buf) - 1) / 2;
		mysql_real_escape_string(&dbwrite.handle, buf, newval, valsz);
		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), ", %s = '%s'", newparam, buf);
	}
	va_end(ap);
	if ((valsz = strlen (lookup)) * 2 + 1 > sizeof(buf))
		valsz = (sizeof(buf) - 1) / 2;
	mysql_real_escape_string(&dbwrite.handle, buf, lookup, valsz);
	snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " WHERE %s = '%s'", keyfield, buf);

	ast_debug(1, "MySQL RealTime: Update SQL: %s\n", sql);

	/* Execution. */
	if (mysql_real_query(&dbwrite.handle, sql, strlen(sql))) {
		ast_log(LOG_WARNING, "MySQL RealTime: Failed to query database. Check debug for more info.\n");
		ast_debug(1, "MySQL RealTime: Query: %s\n", sql);
		ast_debug(1, "MySQL RealTime: Query Failed because: %s\n", mysql_error(&dbwrite.handle));
		ast_mutex_unlock(&dbwrite.lock);
		return -1;
	}

	numrows = mysql_affected_rows(&dbwrite.handle);
	ast_mutex_unlock(&dbwrite.lock);

	ast_debug(1, "MySQL RealTime: Updated %llu rows on table: %s\n", numrows, table);

	/* From http://dev.mysql.com/doc/mysql/en/mysql-affected-rows.html
	 * An integer greater than zero indicates the number of rows affected
	 * Zero indicates that no records were updated
	 * -1 indicates that the query returned an error (although, if the query failed, it should have been caught above.)
	*/

	return (int)numrows;
}
 
static int store_mysql(const char *database, const char *table, va_list ap)
{
	my_ulonglong insertid;
	char sql[512];
	char buf[511]; /* Keep this size uneven as it is 2n+1. */
	char fields[512];
	char values[512];
	char dbname[50] = "";
	int valsz;
	const char *newparam, *newval;

	if (!table) {
		ast_log(LOG_WARNING, "MySQL RealTime: No table specified.\n");
		return -1;
	}
	/* Get the first parameter and first value in our list of passed paramater/value pairs */
	newparam = va_arg(ap, const char *);
	newval = va_arg(ap, const char *);
	if (!newparam || !newval)  {
		ast_log(LOG_WARNING, "MySQL RealTime: Realtime storage requires at least 1 parameter and 1 value to search on.\n");
		return -1;
	}
	/* Must connect to the server before anything else, as the escape function requires the mysql handle. */
	ast_mutex_lock(&dbwrite.lock);
	if (!mysql_reconnect(&dbwrite)) {
		ast_mutex_unlock(&dbwrite.lock);
		return -1;
	}
	/* Create the first part of the query using the first parameter/value pairs we just extracted
		If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat */
	if ((valsz = strlen(newval)) * 2 + 1 > sizeof(buf))
		valsz = (sizeof(buf) - 1) / 2;
	if (newval) {
		mysql_real_escape_string(&dbwrite.handle, buf, newval, valsz);
	} else { 
		buf[0] = '\0';
	}
	snprintf(fields, sizeof(fields), "%s", newparam);
	snprintf(values, sizeof(values), "'%s'", buf);
	while ((newparam = va_arg(ap, const char *))) {
		newval = va_arg(ap, const char *);
		if (newval) {
			if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
				valsz = (sizeof(buf) - 1) / 2;
			mysql_real_escape_string(&dbwrite.handle, buf, newval, valsz);
		} else {
			buf[0] = '\0';
		}
		snprintf(fields + strlen(fields), sizeof(fields), ", %s", newparam);
		snprintf(values + strlen(values), sizeof(values), ", '%s'", buf);
	}
	va_end(ap);
	if (strcmp(database, dbwrite.name)) {
		snprintf(dbname, sizeof(dbname), "%s.", database);
	}
	snprintf(sql, sizeof(sql), "INSERT into %s%s (%s) values (%s)", dbname, table, fields, values);
	ast_debug(1, "MySQL RealTime: Insert SQL: %s\n", sql);

	/* Execution. */
	if (mysql_real_query(&dbwrite.handle, sql, strlen(sql))) {
		ast_log(LOG_WARNING, "MySQL RealTime: Failed to query database. Check debug for more info.\n");
		ast_debug(1, "MySQL RealTime: Query: %s\n", sql);
		ast_debug(1, "MySQL RealTime: Query Failed because: %s\n", mysql_error(&dbwrite.handle));
		ast_mutex_unlock(&dbwrite.lock);
		return -1;
	}

	insertid = mysql_insert_id(&dbwrite.handle);
	ast_mutex_unlock(&dbwrite.lock);

	ast_debug(1, "MySQL RealTime: row inserted on table: %s, id: %llu\n", table, insertid);

	/* From http://dev.mysql.com/doc/mysql/en/mysql-affected-rows.html
	 * An integer greater than zero indicates the number of rows affected
	 * Zero indicates that no records were updated
	 * -1 indicates that the query returned an error (although, if the query failed, it should have been caught above.)
	*/
	return (int)insertid;
}

static int destroy_mysql(const char *database, const char *table, const char *keyfield, const char *lookup, va_list ap)
{
	my_ulonglong numrows;
	char sql[512];
	char buf[511]; /* Keep this size uneven as it is 2n+1. */
	int valsz;
	char dbname[50] = "";
	const char *newparam, *newval;

	if (!table) {
		ast_log(LOG_WARNING, "MySQL RealTime: No table specified.\n");
		return -1;
	}

	/* Get the first parameter and first value in our list of passed paramater/value pairs */
	/* newparam = va_arg(ap, const char *);
	newval = va_arg(ap, const char *);*/
	if (ast_strlen_zero(keyfield) || ast_strlen_zero(lookup))  {
		ast_log(LOG_WARNING, "MySQL RealTime: Realtime destroying requires at least 1 parameter and 1 value to search on.\n");
		return -1;
	}

	/* Must connect to the server before anything else, as the escape function requires the mysql handle. */
	ast_mutex_lock(&dbwrite.lock);
	if (!mysql_reconnect(&dbwrite)) {
		ast_mutex_unlock(&dbwrite.lock);
		return -1;
	}

	/* Create the first part of the query using the first parameter/value pairs we just extracted
	   If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat */

	if ((valsz = strlen (lookup)) * 2 + 1 > sizeof(buf))
		valsz = (sizeof(buf) - 1) / 2;
	mysql_real_escape_string(&dbwrite.handle, buf, lookup, valsz);
	if (strcmp(database, dbwrite.name)) {
		snprintf(dbname, sizeof(dbname), "%s.", database);
	}
	snprintf(sql, sizeof(sql), "DELETE FROM %s%s WHERE %s = '%s'", dbname, table, keyfield, buf);
	while ((newparam = va_arg(ap, const char *))) {
		newval = va_arg(ap, const char *);
		if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
			valsz = (sizeof(buf) - 1) / 2;
		mysql_real_escape_string(&dbwrite.handle, buf, newval, valsz);
		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " AND %s = '%s'", newparam, buf);
	}
	va_end(ap);

	ast_debug(1, "MySQL RealTime: Delete SQL: %s\n", sql);

	/* Execution. */
	if (mysql_real_query(&dbwrite.handle, sql, strlen(sql))) {
		ast_log(LOG_WARNING, "MySQL RealTime: Failed to query database. Check debug for more info.\n");
		ast_debug(1, "MySQL RealTime: Query: %s\n", sql);
		ast_debug(1, "MySQL RealTime: Query Failed because: %s\n", mysql_error(&dbwrite.handle));
		ast_mutex_unlock(&dbwrite.lock);
		return -1;
	}

	numrows = mysql_affected_rows(&dbwrite.handle);
	ast_mutex_unlock(&dbwrite.lock);

	ast_debug(1, "MySQL RealTime: Deleted %llu rows on table: %s\n", numrows, table);

	/* From http://dev.mysql.com/doc/mysql/en/mysql-affected-rows.html
	 * An integer greater than zero indicates the number of rows affected
	 * Zero indicates that no records were updated
	 * -1 indicates that the query returned an error (although, if the query failed, it should have been caught above.)
	*/

	return (int)numrows;
}
 
static struct ast_config *config_mysql(const char *database, const char *table, const char *file, struct ast_config *cfg, struct ast_flags config_flags, const char *unused, const char *who_asked)
{
	MYSQL_RES *result;
	MYSQL_ROW row;
	my_ulonglong num_rows;
	struct ast_variable *new_v;
	struct ast_category *cur_cat;
	char sql[250] = "";
	char last[80] = "";
	char dbname[50] = "";
	int last_cat_metric = 0;

	ast_clear_flag(&config_flags, CONFIG_FLAG_FILEUNCHANGED);
	last[0] = '\0';

	if (!file || !strcmp(file, RES_CONFIG_MYSQL_CONF)) {
		ast_log(LOG_WARNING, "MySQL RealTime: Cannot configure myself.\n");
		return NULL;
	}

	ast_mutex_lock(&dbread.lock);
	if (!mysql_reconnect(&dbread)) {
		ast_mutex_unlock(&dbread.lock);
		return NULL;
	}

	if (strcmp(database, dbread.name)) {
		snprintf(dbname, sizeof(dbname), "%s.", database);
	}
	snprintf(sql, sizeof(sql), "SELECT category, var_name, var_val, cat_metric FROM %s%s WHERE filename='%s' and commented=0 ORDER BY filename, cat_metric desc, var_metric asc, category, var_name, var_val, id", dbname, table, file);

	ast_debug(1, "MySQL RealTime: Static SQL: %s\n", sql);

	if (mysql_real_query(&dbread.handle, sql, strlen(sql))) {
		ast_log(LOG_WARNING, "MySQL RealTime: Failed to query database. Check debug for more info.\n");
		ast_debug(1, "MySQL RealTime: Query: %s\n", sql);
		ast_debug(1, "MySQL RealTime: Query Failed because: %s\n", mysql_error(&dbread.handle));
		ast_mutex_unlock(&dbread.lock);
		return NULL;
	}

	if ((result = mysql_store_result(&dbread.handle))) {
		num_rows = mysql_num_rows(result);
		ast_debug(1, "MySQL RealTime: Found %llu rows.\n", num_rows);

		/* There might exist a better way to access the column names other than counting,
                   but I believe that would require another loop that we don't need. */

		while ((row = mysql_fetch_row(result))) {
			if (!strcmp(row[1], "#include")) {
				if (!ast_config_internal_load(row[2], cfg, config_flags, "", who_asked)) {
					mysql_free_result(result);
					ast_mutex_unlock(&dbread.lock);
					ast_config_destroy(cfg);
					return NULL;
				}
				continue;
			}

			if (strcmp(last, row[0]) || last_cat_metric != atoi(row[3])) {
				cur_cat = ast_category_new(row[0], "", -1);
				if (!cur_cat) {
					ast_log(LOG_WARNING, "Out of memory!\n");
					break;
				}
				strcpy(last, row[0]);
				last_cat_metric = atoi(row[3]);
				ast_category_append(cfg, cur_cat);
			}
			new_v = ast_variable_new(row[1], row[2], "");
			if (cur_cat)
				ast_variable_append(cur_cat, new_v);
		}
	} else {
		ast_log(LOG_WARNING, "MySQL RealTime: Could not find config '%s' in database.\n", file);
	}

	mysql_free_result(result);
	ast_mutex_unlock(&dbread.lock);

	return cfg;
}

static struct ast_config_engine mysql_engine = {
	.name = "mysql",
	.load_func = config_mysql,
	.realtime_func = realtime_mysql,
	.realtime_multi_func = realtime_multi_mysql,
	.store_func = store_mysql,
	.destroy_func = destroy_mysql,
	.update_func = update_mysql
};

static int load_module(void)
{
	int writeokay = 0;
	parse_config();

	ast_mutex_init(&dbread.lock);
	ast_mutex_init(&dbwrite.lock);

	if (!mysql_reconnect(&dbwrite))
		ast_log(LOG_WARNING, "MySQL RealTime: Couldn't establish write connection: %s\n", mysql_error(&dbwrite.handle));
	else
		writeokay = 1;

	if (!mysql_reconnect(&dbread)) {
		if (!writeokay)
			ast_log(LOG_WARNING, "MySQL RealTime: Couldn't establish read connection, either: %s\n", mysql_error(&dbread.handle));
	}

	ast_config_engine_register(&mysql_engine);
	if (option_verbose > 1)
		ast_verbose(VERBOSE_PREFIX_2 "MySQL RealTime driver loaded.\n");
	ast_cli_register_multiple(cli_realtime_mysql_status, sizeof(cli_realtime_mysql_status) / sizeof(struct ast_cli_entry));

	return 0;
}

static int unload_module(void)
{
	ast_cli_unregister_multiple(cli_realtime_mysql_status, sizeof(cli_realtime_mysql_status) / sizeof(struct ast_cli_entry));
	ast_config_engine_deregister(&mysql_engine);
	if (option_verbose > 1)
		ast_verbose(VERBOSE_PREFIX_2 "MySQL RealTime unloaded.\n");

	ast_module_user_hangup_all();

	usleep(1);
	ast_mutex_destroy(&dbread.lock);
	ast_mutex_destroy(&dbwrite.lock);

	mysql_close(&dbwrite.handle);
	mysql_close(&dbread.handle);

	return 0;
}

static int reload(void)
{
	int writeokay = 0;

	ast_mutex_lock(&dbwrite.lock);
	ast_mutex_lock(&dbread.lock);
	mysql_close(&dbwrite.handle);
	mysql_close(&dbread.handle);
	parse_config();

	if (!mysql_reconnect(&dbwrite))
		ast_log(LOG_WARNING, "MySQL RealTime: Cannot connect write handle: %s\n", mysql_error(&dbwrite.handle));
	else
		writeokay = 1;

	if (!mysql_reconnect(&dbread)) {
		if (!writeokay)
			ast_log(LOG_WARNING, "MySQL RealTime: Cannot Connect read handle, either: %s\n", mysql_error(&dbread.handle));
	}
	if (option_verbose > 1)
		ast_verbose(VERBOSE_PREFIX_2 "MySQL RealTime reloaded.\n");

	/* Done reloading. Release lock so others can now use driver. */
	ast_mutex_unlock(&dbread.lock);
	ast_mutex_unlock(&dbwrite.lock);

	return 0;
}

static int set_defaults(struct mysql_conn *conn)
{
	ast_copy_string(conn->user, "asterisk", sizeof(conn->user));
	ast_copy_string(conn->pass, "asterisk", sizeof(conn->pass));
	ast_copy_string(conn->host, "", sizeof(conn->host));
	ast_copy_string(conn->name, "asterisk", sizeof(conn->name));
	conn->port = 3306;
	ast_copy_string(conn->sock, "/tmp/mysql.sock", sizeof(conn->sock));

	if (!ast_strlen_zero(conn->host)) {
		ast_debug(1, "MySQL RealTime host: %s\n", conn->host);
		ast_debug(1, "MySQL RealTime port: %i\n", conn->port);
	} else
		ast_debug(1, "MySQL RealTime socket: %s\n", conn->sock);
	ast_debug(1, "MySQL RealTime database name: %s\n", conn->name);
	ast_debug(1, "MySQL RealTime user: %s\n", conn->user);
	ast_debug(1, "MySQL RealTime password: %s\n", conn->pass);

	return 0;
}

static int parse_config(void)
{
	struct ast_config *config = NULL;
	struct ast_flags config_flags = { 0 };
	const char *catg = "write";
	int haswriteconfig = 0;

	if (!(config = ast_config_load(RES_CONFIG_MYSQL_CONF, config_flags))) {
		set_defaults(&dbwrite);
		set_defaults(&dbread);
		return 0;
	}

	if (!ast_category_exist(config, catg))
		catg = "general";
	if (!ast_category_exist(config, catg)) {
		ast_log(LOG_WARNING, "No configuration set for writing to the MySQL database. Using defaults.\n");
		set_defaults(&dbwrite);
	} else
		haswriteconfig = load_mysql_config(config, catg, &dbwrite);
	
	if (!ast_category_exist(config, "read")) {
		if (haswriteconfig) {
			/* Copy from write config */
			ast_copy_string(dbread.host, dbwrite.host, sizeof(dbread.host));
			ast_copy_string(dbread.sock, dbwrite.sock, sizeof(dbread.sock));
			ast_copy_string(dbread.name, dbwrite.name, sizeof(dbread.name));
			ast_copy_string(dbread.user, dbwrite.user, sizeof(dbread.user));
			ast_copy_string(dbread.pass, dbwrite.pass, sizeof(dbread.pass));
			dbread.port = dbwrite.port;
		} else {
			ast_log(LOG_WARNING, "It's kind of silly to be loading res_config_mysql.so if there's no configuration settings. Let's use the defaults!\n");
			set_defaults(&dbread);
		}
	} else
		load_mysql_config(config, "read", &dbread);

	ast_config_destroy(config);

	return 0;
}

static int load_mysql_config(struct ast_config *config, const char *category, struct mysql_conn *conn)
{
	const char *s;

	if (!(s = ast_variable_retrieve(config, category, "dbuser"))) {
		ast_log(LOG_WARNING, "MySQL RealTime: No database user found, using 'asterisk' as default.\n");
		s = "asterisk";
	}
	ast_copy_string(conn->user, s, sizeof(conn->user));

	if (!(s = ast_variable_retrieve(config, category, "dbpass"))) {
		ast_log(LOG_WARNING, "MySQL RealTime: No database password found, using 'asterisk' as default.\n");
		s = "asterisk";
	}
	ast_copy_string(conn->pass, s, sizeof(conn->pass));

	if (!(s = ast_variable_retrieve(config, category, "dbhost"))) {
		ast_log(LOG_WARNING, "MySQL RealTime: No database host found, using localhost via socket.\n");
		s = "";
	}
	ast_copy_string(conn->host, s, sizeof(conn->host));

	if (!(s = ast_variable_retrieve(config, category, "dbname"))) {
		ast_log(LOG_WARNING, "MySQL RealTime: No database name found, using 'asterisk' as default.\n");
		s = "asterisk";
	}
	ast_copy_string(conn->name, s, sizeof(conn->name) - 1);

	if (!(s = ast_variable_retrieve(config, category, "dbport"))) {
		ast_log(LOG_WARNING, "MySQL RealTime: No database port found, using 3306 as default.\n");
		conn->port = 3306;
	} else
		conn->port = atoi(s);

	if (!(s = ast_variable_retrieve(config, "general", "dbsock"))) {
		if (ast_strlen_zero(conn->host)) {
			ast_log(LOG_WARNING, "MySQL RealTime: No database socket found, using '/tmp/mysql.sock' as default.\n");
			ast_copy_string(conn->sock, "/tmp/mysql.sock", sizeof(conn->sock));
		}
	} else
		ast_copy_string(conn->sock, s, sizeof(conn->sock));

	if (!ast_strlen_zero(conn->host)) {
		ast_debug(1, "MySQL RealTime host: %s\n", conn->host);
		ast_debug(1, "MySQL RealTime port: %i\n", conn->port);
	} else
		ast_debug(1, "MySQL RealTime socket: %s\n", conn->sock);
	ast_debug(1, "MySQL RealTime database name: %s\n", conn->name);
	ast_debug(1, "MySQL RealTime user: %s\n", conn->user);
	ast_debug(1, "MySQL RealTime password: %s\n", conn->pass);

	return 1;
}

static int mysql_reconnect(struct mysql_conn *conn)
{
#ifdef MYSQL_OPT_RECONNECT
	my_bool trueval = 1;
#endif

	/* mutex lock should have been locked before calling this function. */

reconnect_tryagain:
	if ((!conn->connected) && (!ast_strlen_zero(conn->host) || conn->sock) && !ast_strlen_zero(conn->user) && !ast_strlen_zero(conn->name)) {
		if (!mysql_init(&conn->handle)) {
			ast_log(LOG_WARNING, "MySQL RealTime: Insufficient memory to allocate MySQL resource.\n");
			conn->connected = 0;
			return 0;
		}
		if (mysql_real_connect(&conn->handle, conn->host, conn->user, conn->pass, conn->name, conn->port, conn->sock, 0)) {
#ifdef MYSQL_OPT_RECONNECT
			/* The default is no longer to automatically reconnect on failure,
			 * (as of 5.0.3) so we have to set that option here. */
			mysql_options(&conn->handle, MYSQL_OPT_RECONNECT, &trueval);
#endif
			ast_debug(1, "MySQL RealTime: Successfully connected to database.\n");
			conn->connected = 1;
			conn->connect_time = time(NULL);
			return 1;
		} else {
			ast_log(LOG_ERROR, "MySQL RealTime: Failed to connect database server %s on %s (err %d). Check debug for more info.\n", conn->name, !ast_strlen_zero(conn->host) ? conn->host : conn->sock, mysql_errno(&conn->handle));
			ast_debug(1, "MySQL RealTime: Cannot Connect (%d): %s\n", mysql_errno(&conn->handle), mysql_error(&conn->handle));
			conn->connected = 0;
			return 0;
		}
	} else {
		/* MySQL likes to return an error, even if it reconnects successfully.
		 * So the postman pings twice. */
		if (mysql_ping(&conn->handle) != 0 && mysql_ping(&conn->handle) != 0) {
			conn->connected = 0;
			ast_log(LOG_ERROR, "MySQL RealTime: Ping failed (%d).  Trying an explicit reconnect.\n", mysql_errno(&conn->handle));
			ast_debug(1, "MySQL RealTime: Server Error (%d): %s\n", mysql_errno(&conn->handle), mysql_error(&conn->handle));
			goto reconnect_tryagain;
		}

		conn->connected = 1;
		conn->connect_time = time(NULL);

		if (mysql_select_db(&conn->handle, conn->name) != 0) {
			ast_log(LOG_WARNING, "MySQL RealTime: Unable to select database: %s. Still Connected (%u) - %s.\n", conn->name, mysql_errno(&conn->handle), mysql_error(&conn->handle));
			return 0;
		}

		ast_debug(1, "MySQL RealTime: Connection okay.\n");
		return 1;
	}
}

static char *handle_cli_realtime_mysql_status(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	char status[256], status2[100] = "";
	int ctime = 0;

	switch (cmd) {
	case CLI_INIT:
		e->command = "realtime mysql status";
		e->usage =
			"Usage: realtime mysql status\n"
			"       Shows connection information for the MySQL RealTime driver\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3)
		return CLI_SHOWUSAGE;

	ctime = time(NULL) - dbread.connect_time;

	if (mysql_reconnect(&dbread)) {
		if (!ast_strlen_zero(dbread.host))
			snprintf(status, sizeof(status), "Connected to %s@%s, port %d", dbread.name, dbread.host, dbread.port);
		else
			snprintf(status, sizeof(status), "Connected to %s on socket file %s", dbread.name, dbread.sock);

		if (!ast_strlen_zero(dbread.user))
			snprintf(status2, sizeof(status2), " with username %s", dbread.user);

		if (ctime > 31536000)
			ast_cli(a->fd, "%s%s for %.1f years.\n", status, status2, (double)ctime / 31536000.0);
		else if (ctime > 86400 * 30)
			ast_cli(a->fd, "%s%s for %d days.\n", status, status2, ctime / 86400);
		else if (ctime > 86400)
			ast_cli(a->fd, "%s%s for %d days, %d hours.\n", status, status2, ctime / 86400, (ctime % 86400) / 3600);
		else if (ctime > 3600)
			ast_cli(a->fd, "%s%s for %d hours, %d minutes.\n", status, status2, ctime / 3600, (ctime % 3600) / 60);
		else if (ctime > 60)
			ast_cli(a->fd, "%s%s for %d minutes.\n", status, status2, ctime / 60);
		else
			ast_cli(a->fd, "%s%s for %d seconds.\n", status, status2, ctime);

		return CLI_SUCCESS;
	} else
		return CLI_FAILURE;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "MySQL RealTime Configuration Driver",
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		);

